// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Code generated by "stringer -type=pmpResultCode -trimprefix=pmpCode"; DO NOT EDIT.

package portmapper

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[pmpCodeOK-0]
	_ = x[pmpCodeUnsupportedVersion-1]
	_ = x[pmpCodeNotAuthorized-2]
	_ = x[pmpCodeNetworkFailure-3]
	_ = x[pmpCodeOutOfResources-4]
	_ = x[pmpCodeUnsupportedOpcode-5]
}

const _pmpResultCode_name = "OKUnsupportedVersionNotAuthorizedNetworkFailureOutOfResourcesUnsupportedOpcode"

var _pmpResultCode_index = [...]uint8{0, 2, 20, 33, 47, 61, 78}

func (i pmpResultCode) String() string {
	idx := int(i) - 0
	if i < 0 || idx >= len(_pmpResultCode_index)-1 {
		return "pmpResultCode(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _pmpResultCode_name[_pmpResultCode_index[idx]:_pmpResultCode_index[idx+1]]
}
