// Code generated by "stringer -type Label -trimprefix Label"; DO NOT EDIT.

package sockstats

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[LabelControlClientAuto-0]
	_ = x[LabelControlClientDialer-1]
	_ = x[LabelDERPHTTPClient-2]
	_ = x[LabelLogtailLogger-3]
	_ = x[LabelDNSForwarderDoH-4]
	_ = x[LabelDNSForwarderUDP-5]
	_ = x[LabelNetcheckClient-6]
	_ = x[LabelPortmapperClient-7]
	_ = x[LabelMagicsockConnUDP4-8]
	_ = x[LabelMagicsockConnUDP6-9]
	_ = x[LabelNetlogLogger-10]
	_ = x[LabelSockstatlogLogger-11]
	_ = x[LabelDNSForwarderTCP-12]
}

const _Label_name = "ControlClientAutoControlClientDialerDERPHTTPClientLogtailLoggerDNSForwarderDoHDNSForwarderUDPNetcheckClientPortmapperClientMagicsockConnUDP4MagicsockConnUDP6NetlogLoggerSockstatlogLoggerDNSForwarderTCP"

var _Label_index = [...]uint8{0, 17, 36, 50, 63, 78, 93, 107, 123, 140, 157, 169, 186, 201}

func (i Label) String() string {
	idx := int(i) - 0
	if i < 0 || idx >= len(_Label_index)-1 {
		return "Label(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _Label_name[_Label_index[idx]:_Label_index[idx+1]]
}
